searchNodes=[{"doc":"A tool for building in-order GenStage topologies for EventStoreDB Beeline provides a Broadway-like experience for building GenStage topologies for consuming streams of events from EventStoreDB in-order, usually one event at a time. Beeline aims to close over the supervision and basic configuration of the producer(s), as well as some of the run-of-the-mill procedure done in the GenStage.init/1 callback of the consumer such as linking the producer process(es). The Beeline Topology Beeline creates a topology of GenStage, GenServer, and Supervisor processes. This topology looks like the following: Supervisor ├── HealthChecker* └── StageSupervisor ├── Producer* └── Consumer Let's break these down from the bottom up: &quot;Consumer&quot; - the GenStage consumer module which invokes Beeline.start_link/2 , handles events, and increments stream positions. &quot;Producer*&quot; - one or more GenStage producers which feed the consumer. These producers are declared with the :producers key and may either be Kelvin.InOrderSubscription , Volley.InOrderSubscription , or Beeline.DummyProducer producer modules. &quot;StageSupervisor&quot; - a supervisor for the GenStage pipeline. This supervisor has a :transient restart strategy so that if the GenStage pipeline halts on an event it cannot handle, the StageSupervisor supervision tree is brought down but not the entire supervision tree. This behavior is desirable so that the health-checker process can continue reading the stream positions and so that an operator can perform any necessary manual intervention on the crashed supervision tree (for example, skipping the failure event). &quot;HealthChecker*&quot; - a GenServer which periodically polls the stream positions of a producer. There is one health checker process per producer. &quot;Supervisor&quot; - a top-level supervisor. This supervisor has a :permanent restart strategy. See the start_link/2 documentation for a full configuration reference and examples.","ref":"Beeline.html","title":"Beeline","type":"module"},{"doc":"Wraps an event in a subscription event packet This can be useful for building test events to pass through the dummy producer.","ref":"Beeline.html#as_subscription_event/2","title":"Beeline.as_subscription_event/2","type":"function"},{"doc":"Decodes the body of a subscription event This function performs JSON decoding if necessary and converts maps with string keys into maps keyed by atoms. This This can potentially lead to atom exhaustion, but the allowed atom count is quite high and usually this concern is only theoretical. Examples @impl GenStage def handle_events ( [ subscription_event ] , _from , state ) do event = Beeline . decode_event ( subscription_event ) # ..","ref":"Beeline.html#decode_event/1","title":"Beeline.decode_event/1","type":"function"},{"doc":"Determines which producer emitted the subscription event This can be useful in order to save stream positions when a consumer is subscribed to multiple producers. Should be used in tandem with stream_position/1 .","ref":"Beeline.html#producer/1","title":"Beeline.producer/1","type":"function"},{"doc":"Restarts the supervision tree of GenStages for the given Beeline topology This can be useful for manual intervention by a human operator in a remote console session, if the GenStage supervision tree crashes and exceeds the retry limits. Examples iex&gt; Beeline . restart_stages ( MyEventHandler ) :ok","ref":"Beeline.html#restart_stages/1","title":"Beeline.restart_stages/1","type":"function"},{"doc":"Starts a Beeline topology Options :name - The GenServer name for the topology. The topology will build on this name, using it as a prefix. :producers ( keyword/0 ) - A list of producers to which the consumer should subscribe. See the &quot;producer options&quot; section below for the schema. :get_stream_position - A function to invoke in order to get the stream position for a producer. This function should be a 1-arity function (anonymous or capture) where the name of the producer is passed as the argument. This option may also be passed as an MFA tuple where the producer name will be prepended to the argument list. If this option is not provided, a default will be fetched with Application.fetch_env!(:beeline, :get_stream_position) . This configuration can be used to set a blanket function for all beelines to use. :auto_subscribe? - A function to invoke to determine whether each producer should subscribe to events as it starts up. The argument passed is the GenServer name of the producer. If this option is not provided, a default will be fetched with Application.fetch_env!(:beeline, :auto_subscribe?) . :subscribe_after - A period in msec after initialization when each producer should query the :auto_subscribe? function. The default value is {Enum, :random, [3000..5000]} . :spawn_health_checkers? - Controls whether the topology should spawn the HealthChecker children. It can be useful to disable this in Mix.env() in [:dev, :test] as the health checker provides little or no value in those environments and can produce many log lines. If this option is left blank, it will be gotten from the application environment defaulting to true with Application.get_env(:beeline, :spawn_health_checkers?, true) . The default value is nil . :health_check_interval - How long the health checker processes should wait between polling the stream positions. Can either be a function (MFA or 0-arity function) or a non-negative integer. The value is treated as milliseconds. The default value is 51000 . :health_check_drift - A noise to add to the interval specified with :health_check_interval . This can be useful to allow that not all producers poll their positions at the same time, which can reduce strain on the stream position store and the EventStoreDB. Can either be a function (MFA or 0-arity function) or a non-negative integer. The value is treated as milliseconds. If a function is provided, it is invoked every time the health checker process attempts to schedule the next poll. The default value is {Enum, :random, [0..10000]} . :test_mode? - Controls whether the topology should start up in test mode. In test mode, any adapters set in producer specs are switched out with the :dummy adapter. If this option is left blank, it will be gotten from the application environment defaulting to false with Application.get_env(:beeline, :test_mode?, false) . :context ( term/0 ) - A user-defined data structure which is used as the initial state of the GenStage consumer process. The default value is nil . :acceptable_behind_by ( non_neg_integer/0 ) - If listener is behind stream but by this number of events it is considered ok. If in next check it is still behind in this range but didn't process any event since last check, listener is considered stale or &quot;behind&quot;. The default value is 5 . Producer options :adapter - Required. The adapter module to use for creating the producer. Use :kelvin for EventStoreDB v3-5, :volley for EventStoreDB v20+, and :dummy for test cases. :stream_name ( String.t/0 ) - Required. The name of the EventStoreDB stream to which this producer should subscribe for events. :connection ( atom/0 ) - Required. The module to use as a connection to form the subscription. When the :adapter option is :kelvin , this should be an Extreme client module. When the adapter is :volley , it should be a Spear.Client module. :name ( atom/0 ) - The full GenServer name to use for this producer. When this option is not provided, the name will be a formula of the name of the consumer and the key in the keyword list of producers. The default value is nil . :max_demand ( pos_integer/0 ) - The maximum number of events the consumer is allowed to request from this producer. This option can be configured to allow batch processing. The default value is 1 . :min_demand ( non_neg_integer/0 ) - The minimum number of events the consumer can request at a time from this producer. The default value is 0 . Examples defmodule MyEventHandler do use Beeline def start_link ( _opts ) do Beeline . start_link ( MyEventHandler , name : MyEventHandler , producers : [ default : [ name : MyEventHandler.EventListener , stream_name : &quot;$ce-BoundedContext.AggregateName&quot; , connection : MyEventHandler.EventStoreDBConnection , adapter : :kelvin ] ] ) end # .. GenStage callbacks @impl GenStage def handle_events ( [ subscription_event ] , _from , state ) do # .. handle the events one-by-one .. { :noreply , [ ] , state } end end","ref":"Beeline.html#start_link/2","title":"Beeline.start_link/2","type":"function"},{"doc":"Determines the stream position of the subscription event This function prefers link stream positions if available. This means that if the subscription from which the event is emitted is reading a projected stream such as a category stream, the returned stream position will reflect the position in the projected stream instead of the origin stream. Examples @impl GenStage def handle_events ( [ subscription_event ] , _from , state ) do # consume the events MyApp.Repo . transaction ( fn -&gt; # save some state producer = Beeline . producer ( subscription_event ) stream_position = Beeline . stream_position ( subscription_event ) MyApp.StreamPosition . persist ( producer , stream_position ) end ) end","ref":"Beeline.html#stream_position/1","title":"Beeline.stream_position/1","type":"function"},{"doc":"Gives a set of events to a topology's dummy producer This function can be used to test running events through a topology. If there are multiple producers, one is picked at random.","ref":"Beeline.html#test_events/2","title":"Beeline.test_events/2","type":"function"},{"doc":"A GenStage producer built for testing beeline topologies This producer can be used either by setting the producer option :adapter to :dummy or by setting the :test_mode? option to true . This producer emits events which arrive via GenServer.cast/2 . Use Beeline.test_events/2 to send events to a topology's dummy producer.","ref":"Beeline.DummyProducer.html","title":"Beeline.DummyProducer","type":"module"},{"doc":"A GenServer which periodically polls a producer's stream positions and process This GenServer emits :telemetry measurements which serve as an interface for exporting this health-check information to an external monitoring service. Beeline.HealthChecker.Logger is included as a reasonable default consumer for this telemetry. You may wish to export this telemetry to another system such as appsignal or grafana in order to create alerts when processors fall behind. Telemetry [:beeline, :health_check, :stop] - dispatched by each HealthChecker process after polling the producer's position and process information Measurement: %{duration: native_time} - the time taken to perform the position and process checks Metadata, a map with the following keys: :producer (module) - the producer module being measured :alive? (boolean) - whether the producer process is alive :stream_name (string) - the EventStoreDB stream name from which the producer reads :hostname (string) - the hostname of the machine on which the health checker process is being run :interval (integer) - the milliseconds the health checker process has waited (minus drift) since the last poll :drift (integer) - the milliseconds used for drifting the interval for the last poll :measurement_time (UTC datetime) - the time when the poll started :prior_position (integer) - the :current_position from the last poll :current_position (integer) - the current stream position of the producer :head_position (integer) - the stream position of the head (the latest event) of the EventStoreDB stream :auto_subscribe (boolean) - the value of the producer's :auto_subscribe? flag at the time of polling","ref":"Beeline.HealthChecker.html","title":"Beeline.HealthChecker","type":"module"},{"doc":"A Task to attach a logger exporter for health-checker telemetry Attaches a telemetry handler which writes out how far a producer is behind the head of the stream to the logger. This task can be started in a supervision tree such as an app's application supervision tree: def start ( _type , _args ) do children = [ Beeline.HealthChecker.Logger , MyApp.MyBeelineTopology ] opts = [ strategy : :one_for_one , name : MyApp.Supervisor ] Supervisor . start_link ( children , opts ) end The log messages are 'debug' level in a format of the producer name concatenated with &quot;is up-to-date.&quot; if the producer's current stream position matches the latest available stream position (within acceptable_behind_by ). When producer is falling behind and falling behind more then in a previous check, it is logged in warn level with &quot;is behind: n events.&quot; and &quot;is behind more: n events&quot; with n being the number of events producer is behind head of stream. Once producer is behind the stream less then in a previous check and when it catches up with it, it is logged in info level as is behind but catching up: n events and is caught up. respectively. The log messages also include metadata fields :event_listener - the name of the producer - and delta : the number of events by which the producer is behind.","ref":"Beeline.HealthChecker.Logger.html","title":"Beeline.HealthChecker.Logger","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. arg is passed as the argument to Task.start_link/1 in the :start field of the spec. For more information, see the Supervisor module, the Supervisor.child_spec/2 function and the Supervisor.child_spec/0 type.","ref":"Beeline.HealthChecker.Logger.html#child_spec/1","title":"Beeline.HealthChecker.Logger.child_spec/1","type":"function"},{"doc":"All notable changes to this project will be documented in this file. The format is based on Keep a Changelog , and this project adheres to Semantic Versioning .","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Changed Log everything health check related with debug level if auto_subscribe is false .","ref":"changelog.html#1-1-4-2023-09-18","title":"Changelog - 1.1.4 - 2023-09-18","type":"extras"},{"doc":"Added Added previous_status to metadata of telemetry event.","ref":"changelog.html#1-1-3-2023-07-21","title":"Changelog - 1.1.3 - 2023-07-21","type":"extras"},{"doc":"Changed More accurate health check monitoring and logging","ref":"changelog.html#1-1-2-2023-07-14","title":"Changelog - 1.1.2 - 2023-07-14","type":"extras"},{"doc":"Changed &quot;caught up&quot; messages logged at debug level","ref":"changelog.html#1-1-1-2023-04-04","title":"Changelog - 1.1.1 - 2023-04-04","type":"extras"},{"doc":"This change adds the ability to use :global and Registry names for the process and topology.","ref":"changelog.html#1-1-0-2022-10-12","title":"Changelog - 1.1.0 - 2022-10-12","type":"extras"},{"doc":"This change is for stability only and does not present any functional changes from v0.3.1.","ref":"changelog.html#1-0-0-2022-05-11","title":"Changelog - 1.0.0 - 2022-05-11","type":"extras"},{"doc":"Fixed Use wrap_function to properly evaluate auto_subscribe value","ref":"changelog.html#0-3-1-2022-05-05","title":"Changelog - 0.3.1 - 2022-05-05","type":"extras"},{"doc":"Added Added auto_subscribe field to telemetry event","ref":"changelog.html#0-3-0-2022-04-14","title":"Changelog - 0.3.0 - 2022-04-14","type":"extras"},{"doc":"Fixed The extreme adapter for fetching the latest event number has been patched to treat the extreme connection as a process instead of a module this allows one to use connections which are not modules Fixed whitespace in ArgumentError message for invalid configuration","ref":"changelog.html#0-2-0-2021-07-14","title":"Changelog - 0.2.0 - 2021-07-14","type":"extras"},{"doc":"Added Initial implementation published to Hex.pm","ref":"changelog.html#0-1-0-2021-06-15","title":"Changelog - 0.1.0 - 2021-06-15","type":"extras"},{"doc":"Added This project was generated by Gaas","ref":"changelog.html#0-0-0-2021-06-09","title":"Changelog - 0.0.0 - 2021-06-09","type":"extras"}]